#include <pspkernel.h>
#include <psploadcore.h>
#include <pspiofilemgr.h>
#include <pspdisplay.h>
#include <pspctrl.h>
#include <stdio.h>
#include <string.h>
#include <psppower.h>
#include <pspsysmem_kernel.h>
#include <psprtc.h>
#include <psphprm.h> 

PSP_MODULE_INFO("Screen Shot: BMP", 0x1000, 1, 0);
PSP_HEAP_SIZE_KB(64);
PSP_MAIN_THREAD_ATTR(0);

#define MAX_THREAD	64
//add by mascii
#define ANALOG_SENS 80
//

typedef unsigned char   BYTE;
typedef unsigned short  WORD;
typedef unsigned long   DWORD;
typedef long            LONG;

typedef struct tagBITMAPFILEHEADER {
        DWORD   bfSize;
        DWORD   bfReserved;
        DWORD   bfOffBits;
} BITMAPFILEHEADER;

typedef struct tagBITMAPINFOHEADER{
	DWORD	biSize;
	LONG	biWidth;
	LONG	biHeight;
	WORD	biPlanes;
	WORD	biBitCount;
	DWORD	biCompression;
	DWORD	biSizeImage;
	LONG	biXPelsPerMeter;
	LONG	biYPelsPerMeter;
	DWORD	biClrUsed;
	DWORD	biClrImportant;
} BITMAPINFOHEADER;

	char file[64], flag = 0, size = 0;
	int x, y, fd, count = 0, thread_count_start, thread_count_now;
	SceUID thread_buf_start[MAX_THREAD], thread_buf_now[MAX_THREAD], myThread;

//add by mascii
	int AX = 0;
	int AY = 0;
//

/***************************************

画面をBMPファイルに保存

****************************************/
void screenshot_BMP(const char* filename){
	
	int x, y;
	scePowerLock(0);
	int fd = sceIoOpen(filename, PSP_O_CREAT|PSP_O_WRONLY|PSP_O_TRUNC, 0777);
	if(fd < 0){
		return;
	}

	int pwidth, pheight, bufferwidth, pixelformat, unk;
	unsigned int* vram32orig;
	unsigned int* vram32;
	unsigned short* vram16;

	//ディスプレイ情報取得
	sceDisplayWaitVblankStart();
	sceDisplayGetMode(&unk, &pwidth, &pheight);
	sceDisplayGetFrameBuf((void*)&vram32orig, &bufferwidth, &pixelformat, &unk);
	int i;
	int pw = pwidth, ph = pheight;

	
	sceKernelSetDdrMemoryProtection((void*)0x88400000, 0x00400000, 0xF);
	int size = pixelformat == 3 ? (bufferwidth*pw*sizeof(u32)) : (bufferwidth*pw*sizeof(u16));
	int id = sceKernelAllocPartitionMemory(5, "scrbuf", 1, size, NULL);
	vram32 = (unsigned int*) sceKernelGetBlockHeadAddr(id);
	if(!vram32) 
	{
		//the pointer is invalid
		//closing and removing the file and returning
		sceIoClose(fd);
		sceIoRemove(filename);
		return;
	}	

		
	for(i = 0;  i < size ; i++)
	{
		*(char*)((int)vram32+i) = *(char*)((int)vram32orig+i);
	}

	vram16 = (unsigned short*) vram32;


			for(x = 0; x < thread_count_now; x++){
			// thread id match 0 or 1
			unsigned char match = 0;
			SceUID tmp_thid = thread_buf_now[x];
			for(y = 0; y < thread_count_start; y++){
				if((tmp_thid == thread_buf_start[y]) || (tmp_thid == myThread)){
					match = 1;
					y = thread_count_start;
				}
			}
			if(thread_count_start == 0) match = 1;
			if(match == 0){
			sceKernelResumeThread(tmp_thid);
		}
	}
		
	//BMPヘッダ作成
	unsigned char bm[2], padding;
	BITMAPFILEHEADER h1;
	BITMAPINFOHEADER h2;
	
	//BM書き込み
	bm[0] = 0x42;
	bm[1] = 0x4D;
	sceIoWrite(fd, bm, 2);

		
	padding = (3*(pw/2) % 4);
	h1.bfSize	= ((24*pw + padding)*ph)/8 + sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER) + 2;
	h1.bfReserved	= 0;
	h1.bfOffBits	= 2 + sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);

	h2.biSize	= sizeof(BITMAPINFOHEADER);
	h2.biPlanes	= 1;
	h2.biBitCount	= 24;
	h2.biCompression 	= 0;
	h2.biWidth	= pw;
	h2.biHeight	= ph;
	h2.biSizeImage	= ((24*pw + padding)*ph)/8;
	h2.biXPelsPerMeter	= 0xEC4;
	h2.biYPelsPerMeter	= 0xEC4;
	h2.biClrUsed		= 0;
	h2.biClrImportant	= 0;

	//ヘッダ書き込み
	sceIoWrite(fd, &h1, sizeof(BITMAPFILEHEADER));
	sceIoWrite(fd, &h2, sizeof(BITMAPINFOHEADER));
	
	
//スクリーンを24bitBMP画像に変換
	SceUID mem = sceKernelAllocPartitionMemory(2, "block", 0, (3*pw + padding)*sizeof(unsigned char), NULL);
	unsigned char *buf = (unsigned char*)sceKernelGetBlockHeadAddr(mem);


	for(x = 0; x < padding; x++){
		buf[3*pw + x] = 0;
	}
	for(y = (ph-1); y >= 0; y--){
		int i;
		for(i = 0, x = 0; x < pw; x++){
			unsigned int color, offset = x + y*bufferwidth;
			unsigned char r = 0, g = 0, b = 0;

			switch (pixelformat) {
				case 0:	// 16-bit RGB 5:6:5
					color = vram16[offset];
					r = (color & 0x1f) << 3; 
					g = ((color >> 5) & 0x3f) << 2 ;
					b = ((color >> 11) & 0x1f) << 3 ;
					break;
				case 1:// 16-bit RGBA 5:5:5:1
					color = vram16[offset];
					r = (color & 0x1f) << 3; 
					g = ((color >> 5) & 0x1f) << 3 ;
					b = ((color >> 10) & 0x1f) << 3 ;
					break;
				case 2:// 16-bit RGBA 4:4:4:4
					color = vram16[offset];
					r = (color & 0xf) << 4; 
					g = ((color >> 4) & 0xf) << 4 ;
					b = ((color >> 8) & 0xf) << 4 ;
					break;
				case 3:// 32-bit RGBA 8:8:8:8
					color = vram32[offset];
					r = color & 0xff; 
					g = (color >> 8) & 0xff;
					b = (color >> 16) & 0xff;
					break;
			}
			buf[i++] = b;
			buf[i++] = g;
			buf[i++] = r;
		}
		sceIoWrite(fd, buf, 3*pw + padding);
	}
	sceKernelFreePartitionMemory(id);
	sceKernelFreePartitionMemory(mem);
	sceIoClose(fd);
	scePowerUnlock(0);
}

int mainThread (SceSize args, void *argp){
	unsigned int paddata_old = 0;
	//add by mascii
	unsigned int paddata_oldlx = 0;
	unsigned int paddata_oldly = 0;
	//
	myThread = sceKernelGetThreadId();
	SceCtrlData paddata;
	pspTime rtime;
	SceUID fd;
	SceUID memid;
	u32 filesize;
	char *buf;
	char line[300];
	u32 pos;
	char *str;
	int x;
	int buttons = 0;
	
	sceKernelDelayThread(10000);

	sceKernelGetThreadmanIdList(SCE_KERNEL_TMID_Thread, thread_buf_start, MAX_THREAD, &thread_count_start);

	// ボタン設定読み込み
    fd = sceIoOpen("ms0:/seplugins/ss_conf.txt",PSP_O_RDONLY,0777);
    if(fd < 0)
    {
		buttons = PSP_CTRL_NOTE;
	}
	else
	{
		filesize = sceIoLseek32(fd, 0, PSP_SEEK_END);
		sceIoLseek32(fd, 0, PSP_SEEK_SET);
		memid = sceKernelAllocPartitionMemory(1, "CONFIG_MEM", PSP_SMEM_Low, filesize, NULL);
		buf = sceKernelGetBlockHeadAddr(memid);
		sceIoRead(fd, buf, filesize);

		buf[filesize] = '\0'; //make the file a null-terminated string
		sceIoClose(fd);

		pos = strcspn(buf,"\n\r#");
		strncpy(line,buf,pos);
		line[pos]= '\0';
		str = strtok(line,"|");
		
		while(str != NULL)
		{
			if(strcmp(str,"SELECT")==0)
				buttons |= PSP_CTRL_SELECT;
			if(strcmp(str,"START")==0)
				buttons |= PSP_CTRL_START;
			if(strcmp(str,"UP")==0)
				buttons |= PSP_CTRL_UP;
			if(strcmp(str,"DOWN")==0)
				buttons |= PSP_CTRL_DOWN;
			if(strcmp(str,"LEFT")==0)
				buttons |= PSP_CTRL_LEFT;
			if(strcmp(str,"RIGHT")==0)
				buttons |= PSP_CTRL_RIGHT;
			if(strcmp(str,"LTRIGGER")==0)
				buttons |= PSP_CTRL_LTRIGGER;
			if(strcmp(str,"RTRIGGER")==0)
				buttons |= PSP_CTRL_RTRIGGER;
			if(strcmp(str,"TRIANGLE")==0)
				buttons |= PSP_CTRL_TRIANGLE;
			if(strcmp(str,"CIRCLE")==0)
				buttons |= PSP_CTRL_CIRCLE;
			if(strcmp(str,"CROSS")==0)
				buttons |= PSP_CTRL_CROSS;
			if(strcmp(str,"SQUARE")==0)
				buttons |= PSP_CTRL_SQUARE;
			if(strcmp(str,"HOME")==0)
				buttons |= PSP_CTRL_HOME;

			if(strcmp(str,"HOLD")==0) //not really a button
				buttons |= PSP_CTRL_HOLD;
			if(strcmp(str,"DISABLE")==0)//hold prevents any buttons from working so it acts as a disable switch
				return PSP_CTRL_HOLD;

			if(strcmp(str,"NOTE")==0)
				buttons |= PSP_CTRL_NOTE;

			if(strcmp(str,"SCREEN")==0)
				buttons |= PSP_CTRL_SCREEN;

			if(strcmp(str,"VOLUP")==0)
				buttons |= PSP_CTRL_VOLUP;

			if(strcmp(str,"VOLDOWN")==0)
				buttons |= PSP_CTRL_VOLDOWN;

			//add by mascii
			if(strcmp(str,"REMOTE")==0)
				buttons |= PSP_CTRL_REMOTE;

			if(strcmp(str,"WLAN_UP")==0)
				buttons |= PSP_CTRL_WLAN_UP;

			if(strcmp(str,"A_UP")==0)
				AY--;

			if(strcmp(str,"A_DOWN")==0)
				AY++;

			if(strcmp(str,"A_LEFT")==0)
				AX--;

			if(strcmp(str,"A_RIGHT")==0)
				AX++;
			//

			str = strtok(NULL,"|");
		}

		//add by mascii
		if(!buttons && !AX && !AY){
			buttons = PSP_CTRL_NOTE;
		}
		//

		sceKernelFreePartitionMemory(memid);
	}
	
	while(1){
		sceCtrlPeekBufferPositive(&paddata, 1);

		//add by mascii
		sceCtrlSetSamplingMode(PSP_CTRL_MODE_ANALOG);
		//

		//mod by mascii
		if((paddata.Buttons != paddata_old) || (paddata.Lx != paddata_oldlx) || (paddata.Ly != paddata_oldly)){
			//スクリーンショット
			//mod by mascii
			if((paddata.Buttons & buttons) == buttons || (!buttons && (AX || AY))){
			if(AX==0 || ( AX==-1 && paddata.Lx < 128 - ANALOG_SENS) || (AX==1 && paddata.Lx > 128 + ANALOG_SENS)){
			if(AY==0 || ( AY==-1 && paddata.Ly < 128 - ANALOG_SENS) || (AY==1 && paddata.Ly > 128 + ANALOG_SENS)){
		
				// IdList Now
				sceKernelGetThreadmanIdList(SCE_KERNEL_TMID_Thread, thread_buf_now, MAX_THREAD, &thread_count_now);

				for(x = 0; x < thread_count_now; x++){
					// thread id match 0 or 1
					unsigned char match = 0;
					SceUID tmp_thid = thread_buf_now[x];
					for(y = 0; y < thread_count_start; y++){
						if((tmp_thid == thread_buf_start[y]) || (tmp_thid == myThread)){
							match = 1;
							y = thread_count_start;
						}
					}
					if(thread_count_start == 0) match = 1;
					if(match == 0){
						sceKernelSuspendThread(tmp_thid);
					}

				}
					
				
				//スクリーンショットの処理
					sceIoMkdir("ms0:/PICTURE",0777);
					sceIoMkdir("ms0:/PICTURE/capture",0777);
				
				sceRtcGetCurrentClockLocalTime(&rtime);
					sprintf(file, "ms0:/PICTURE/capture/%04d%02d%02d%02d%02d%02d.bmp", rtime.year, rtime.month, rtime.day, rtime.hour, rtime.minutes, rtime.seconds);
				
				fd = sceIoOpen(file, PSP_O_RDONLY, 0644);
				sceIoClose(fd);
				screenshot_BMP(file);
				
				// スレッドを再開

				
			}}}
		}
		paddata_old = paddata.Buttons;
		paddata_oldlx = paddata.Lx;
		paddata_oldly = paddata.Ly;
		sceKernelDelayThread(10000);
	}
	return 0;
}

int module_start (SceSize args, void *argp) __attribute__((alias("plugin_entry")));

int plugin_entry (SceSize args, void* argp)
{
	SceUID thid;
	u32 func = 0x80000000 | (u32)mainThread;

	thid = sceKernelCreateThread("ch_main", (void*)func, 0x18, 0x1000, 0, NULL);
	if (thid >= 0) sceKernelStartThread(thid, args, argp);

	return 0;
}
